package com.tibco.utils.scheduler;

import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.quartz.CronExpression;
import org.quartz.CronScheduleBuilder;
import org.quartz.JobBuilder;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.ScheduleBuilder;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.SimpleScheduleBuilder;
import org.quartz.Trigger;
import org.quartz.TriggerBuilder;
import org.quartz.TriggerKey;
import org.quartz.impl.StdSchedulerFactory;
import org.quartz.impl.matchers.GroupMatcher;
import org.slf4j.Logger;

import com.streambase.sb.DataType;
import com.streambase.sb.Schema;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.NullValueException;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.operator.Operator;
import com.streambase.sb.operator.Parameterizable;
import com.streambase.sb.operator.TypecheckException;
import com.tibco.utils.scheduler.job.SchedulerJob;

/**
 * Generated by JDT StreamBase Client Templates (Version: 7.6.5.1608041644).
 *
 * This class is used as a Java Operator in a StreamBase application.
 * One instance will be created for each Java Operator in a StreamBase 
 * application. 
 * <p>
 * Enqueue methods should only be called from processTuple.
 * <p>
 * If there is any state that the operator instance needs to maintain beyond the
 * lifetime of the instance, the setSessionState/getSessionState methods can
 * be used.
 * @see Parameterizable
 * @see Operator
 * For in-depth information on implementing a custom Java Operator, please see
 * "Developing StreamBase Java Operators" in the StreamBase documentation.
 */
public class TasksScheduler extends Operator implements Parameterizable {

	public static final long serialVersionUID = 1479262318667L;
	public static final String JOB_DATA_OPERATOR_INSTANCE = "OPERATOR_INSTANCE";
	public static final String JOB_DATA_SCHEDULE_CONFIGURATION_TUPLE = "JOB_DATA_SCHEDULE_CONFIGURATION_TUPLE";

	private static final String SCHEDULER_GROUP_TRIGGER = "group-trigger";
	private static final String SCHEDULER_GROUP_JOB = "group-job";

	public static final String scheduleConfigurationFieldName = "ScheduleConfiguration";
	public static final String commandFieldName = "command";
	public static final String idFieldName = "id";
	public static final String cronExpressionFieldName = "cronExpression";
	public static final String contextFieldName = "context";
	public static final String repeatWithSecondsIntervalFieldName = "repeatWithSecondsInterval";
	
	private Logger logger;
	private String displayName = "Tasks Scheduler";
	// Local variables
	private int inputPorts = 1;
	private int outputPorts = 1;
	private Schema inputSchema;
	private Schema[] outputSchemas; // caches the Schemas given during init() for use at processTuple()
	
	private Scheduler scheduler;
	
	/**
	 * The constructor is called when the Operator instance is created, but before the Operator 
	 * is connected to the StreamBase application. We recommended that you set the initial input
	 * port and output port count in the constructor by calling setPortHints(inPortCount, outPortCount).
	 * The default is 1 input port, 1 output port. The constructor may also set default values for 
	 * operator parameters. These values will be displayed in StreamBase Studio when a new instance
	 * of this operator is  dragged to the canvas, and serve as the default values for omitted
	 * optional parameters.
	 */
	public TasksScheduler() {
		super();
		logger = getLogger();
		setPortHints(inputPorts, outputPorts);
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());

	}

	@Override
	public URL getIconResource(IconKind iconType) {
		switch (iconType) {
		case CANVAS_OVERLAY_ICON:
			return getClass().getResource("Schedule_clock_48x48.png");
		case PALETTE_ICON_LARGE:
			return getClass().getResource("Schedule_clock_24x24.png");
		case PALETTE_ICON_SMALL:
			return getClass().getResource("Schedule_clock_16x16.png");
		default:
			return null;
		}
	}

	/**
	 * The typecheck method is called after the Operator instance is connected in the StreamBase
	 * application, allowing the Operator to validate its properties. The Operator class may 
	 * change the number of input or output ports by calling the requireInputPortCount(portCount)
	 * method or the setOutputSchema(schema, portNum) method. If the verifyInputPortCount method 
	 * is passed a different number of ports than the Operator currently has, a PortMismatchException
	 * (subtype of TypecheckException) is thrown.
	 */
	public void typecheck() throws TypecheckException {
		// typecheck: require a specific number of input ports
		requireInputPortCount(inputPorts);

		// Ensure that all properties have valid values, and typecheck the input schemas here
		// Only 1 input port
		inputSchema = getInputSchema(0);

		// Check if the schema contains a Tuple named "ScheduleConfiguration"
		Field scheduleTupleField; 
		try {
			scheduleTupleField = inputSchema.getField(scheduleConfigurationFieldName);
		} catch (TupleException e) {
			throw new TypecheckException("The input tuple must have a field ["+scheduleConfigurationFieldName+"] of type [tuple]",e);
		}
		Schema scheduleConfigurationSchema = scheduleTupleField.getSchema();
		
		// Check if the schema has a command, id, cronExpression and/or repeatInterval and context
		Field commandField;
		try{
			commandField = scheduleConfigurationSchema.getField(commandFieldName);
			if(!DataType.STRING.equals(commandField.getDataType())){
				throw new TypecheckException("The field ["+commandFieldName+"] must be of type [string]");
			}
		}catch(TupleException e){
			throw new TypecheckException("The tuple ["+scheduleConfigurationFieldName+"] must have a field ["+commandFieldName+"] of type [string]",e);
		}
		// ID
		Field idField;
		try {
			idField = scheduleConfigurationSchema.getField(idFieldName);
			if(!DataType.STRING.equals(idField.getDataType())){
				throw new TypecheckException("The field ["+idFieldName+"] must be of type [string]");
			}
		} catch (TupleException e) {
			throw new TypecheckException("The tuple ["+scheduleConfigurationFieldName+"] must have a field ["+idFieldName+"] of type [string]",e);
		}
		
		// Both cronExpression and repeatWithSecondsInterval fields
		Field cronExpressionField = null;
		Field repeatWithSecondsIntervalField = null;
		try {
			cronExpressionField = scheduleConfigurationSchema.getField(cronExpressionFieldName);
			if(!DataType.STRING.equals(cronExpressionField.getDataType())){
				throw new TypecheckException("The field ["+cronExpressionFieldName+"] must be of type [string]");
			}
		} catch (TupleException e) {} // Nothing since we want to check if one or the other is here
		try {
			repeatWithSecondsIntervalField = scheduleConfigurationSchema.getField(repeatWithSecondsIntervalFieldName);
			if(!DataType.INT.equals(repeatWithSecondsIntervalField.getDataType())){
				throw new TypecheckException("The field ["+repeatWithSecondsIntervalFieldName+"] must be of type [int]");
			}
		} catch (TupleException e) {} // Nothing since we want to check if one or the other is here
		
		// Create an output schema for our port
		List<Schema.Field> outputSchemaFields = new ArrayList<Schema.Field>();
		outputSchemaFields.add(scheduleTupleField);
		setOutputSchema(0, new Schema(null, outputSchemaFields));

	}

	/**
	 * This method will be called by the StreamBase server for each Tuple given
	 * to the Operator to process. This is the only time an operator should 
	 * enqueue output Tuples.
	 * @param inputPort the input port that the tuple is from (ports are zero based)
	 * @param tuple the tuple from the given input port
	 * @throws StreamBaseException Terminates the application.
	 */
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
		if (inputPort > 0) {
			logger.info("operator skipping tuple at input port" + inputPort);
			return;
		}
		
		Tuple scheduleConfigurationTuple = tuple.getTuple(scheduleConfigurationFieldName);
		if(scheduleConfigurationTuple == null){
			throw new StreamBaseException("The ["+scheduleConfigurationFieldName+"] is null.");
		}
		
		// Confirm if there is a command.
		String command = null;
		try{
			command = scheduleConfigurationTuple.getString(commandFieldName);
		}catch(NullValueException e){}
		
		if(command == null || !("start".equalsIgnoreCase(command) || "stop".equalsIgnoreCase(command))){
			throw new StreamBaseException("the field ["+commandFieldName+"] must be one of [start,stop]");
		}
		
		
		String id = null;
		try{
			id = scheduleConfigurationTuple.getString(idFieldName);
		}catch(NullValueException e){}
		
		if("start".equalsIgnoreCase(command) && id == null){
			throw new StreamBaseException("The field ["+idFieldName+"] cannot be null if the command is \"start\".");
		}
		
		if("start".equalsIgnoreCase(command)){
			ScheduleBuilder<? extends Trigger> schedule = null;
			String cronExpression = null; 
			try{
				cronExpression = scheduleConfigurationTuple.getString(cronExpressionFieldName);
			}catch(NullValueException e){} // Do nothing since the other field could be set.
			
			int repeatInterval = -1;
			try{
				repeatInterval = scheduleConfigurationTuple.getInt(repeatWithSecondsIntervalFieldName);
			}catch(NullValueException e){} // Do nothing since the other field could be set.
			
			if("start".equalsIgnoreCase(command) && cronExpression == null && repeatInterval == -1){
				throw new StreamBaseException("One of ["+cronExpressionFieldName+"] or ["+repeatWithSecondsIntervalFieldName+"] field must be present when the command is \"start\"");
			}else if("start".equalsIgnoreCase(command) && cronExpression != null){
				if(CronExpression.isValidExpression(cronExpression)){
					schedule = CronScheduleBuilder.cronSchedule(cronExpression);
				}else{
					throw new StreamBaseException("["+cronExpression+"] is not a valid Cron Expression.");
				}
			}else if("start".equalsIgnoreCase(command) && repeatInterval >= 0){
				schedule = SimpleScheduleBuilder.repeatSecondlyForever(repeatInterval);
			}
			scheduleJob(scheduleConfigurationTuple, id, schedule);
		}else if("stop".equalsIgnoreCase(command)){
			stopJob(id);
		}
		
	}

	/**
	 * Schedule the job using a unique ID, a schedule and the scheduleConfigurationTuple (as context)
	 * If the job already exists, it will be replaced.
	 * @param scheduleConfigurationTuple
	 * @param id
	 * @param schedule
	 * @throws StreamBaseException
	 */
	private void scheduleJob(Tuple scheduleConfigurationTuple, String id, ScheduleBuilder<? extends Trigger> schedule)
			throws StreamBaseException {
		JobKey jobKey = new JobKey("job-"+id, SCHEDULER_GROUP_JOB);
		JobDetail job = null;
		try {
			job = scheduler.getJobDetail(jobKey);
		} catch (SchedulerException e1) {
			throw new StreamBaseException("Cannot retrieve the job detail using key [job-"+id+"]", e1);
		}
		TriggerKey triggerKey = new TriggerKey("trigger-"+id, SCHEDULER_GROUP_TRIGGER);
		Trigger trigger;
		try {
			trigger = scheduler.getTrigger(triggerKey);
		} catch (SchedulerException e1) {
			throw new StreamBaseException("Cannot retrieve the trigger using key [trigger-"+id+"]", e1);
		}
		
		JobDataMap jobDataMap = new JobDataMap();
		jobDataMap.put(JOB_DATA_OPERATOR_INSTANCE, this);
		jobDataMap.put(JOB_DATA_SCHEDULE_CONFIGURATION_TUPLE, scheduleConfigurationTuple);

		// Trigger the job to run now, and then every 40 seconds
		Trigger newTrigger = TriggerBuilder.newTrigger()
				.withIdentity(triggerKey)
				.startNow()
				.withSchedule(schedule)
				.build();

		JobDetail newJob = JobBuilder.newJob(SchedulerJob.class)
				.withIdentity(jobKey)
				.usingJobData(jobDataMap)
				.storeDurably(true)
				.build();
		if(job != null && trigger != null){
			logger.warn("A job with the same id already exists. It will be replaced with the new value");
			try {
				scheduler.addJob(newJob, true);
			} catch (SchedulerException e) {
				throw new StreamBaseException("Couldn't replace the job.", e);
			}
			try {
				scheduler.rescheduleJob(triggerKey, newTrigger);
			} catch (SchedulerException e) {
				throw new StreamBaseException("Couldn't reschedule the trigger.", e);
			}
		}else{
			// Tell quartz to schedule the job using our trigger
			try {
				scheduler.scheduleJob(newJob, newTrigger);
			} catch (SchedulerException e) {
				throw new StreamBaseException("Couldn't schedule the job.", e);
			}

		}
	}

	/**
	 * Stop a job by ID. If id is null, all jobs are stopped.
	 * @param id
	 */
	private void stopJob(String id) throws StreamBaseException{
		GroupMatcher<JobKey> jobGroupMatcher = GroupMatcher.jobGroupEquals(SCHEDULER_GROUP_JOB);
		GroupMatcher<TriggerKey> triggerGroupMatcher = GroupMatcher.triggerGroupEquals(SCHEDULER_GROUP_TRIGGER);
		if(id == null){
			try {
				for(JobKey jobKey: scheduler.getJobKeys(jobGroupMatcher)){
					try {
						scheduler.deleteJob(jobKey);
					} catch (SchedulerException e) {
						throw new StreamBaseException("Couldn't delete the job for id ["+jobKey.getName()+"]", e);
					}
				}
				for(TriggerKey triggerKey: scheduler.getTriggerKeys(triggerGroupMatcher)){
					try{
						scheduler.unscheduleJob(triggerKey);
					} catch (SchedulerException e) {
						throw new StreamBaseException("Couldn't unschedule the trigger for id ["+triggerKey.getName()+"]", e);
					}
				}
			} catch (SchedulerException e) {
				throw new StreamBaseException("Couldn't retrieve the list of JobKeys", e);
			}
		}else{
			JobKey jobKey = new JobKey("job-"+id, SCHEDULER_GROUP_JOB);
			TriggerKey triggerKey = new TriggerKey("trigger-"+id, SCHEDULER_GROUP_TRIGGER);
			try {
				scheduler.deleteJob(jobKey);
			} catch (SchedulerException e) {
				throw new StreamBaseException("Couldn't delete the job for id ["+id+"]", e);
			}
			try {
				scheduler.unscheduleJob(triggerKey);
			} catch (SchedulerException e) {
				throw new StreamBaseException("Couldn't delete the trigger  for id ["+id+"]", e);
			}
		}
	}
	
	/**
	 * If typecheck succeeds, the init method is called before the StreamBase application
	 * is started. Note that your Operator class is not required to define the init method,
	 * unless (for example) you need to perform initialization of a resource such as a JDBC
	 * pool, if your operator is making JDBC calls. StreamBase Studio does not call this
	 * during authoring.
	 */
	public void init() throws StreamBaseException {
		super.init();
		// for best performance, consider caching input or output Schema.Field objects for
		// use later in processTuple()
		outputSchemas = new Schema[outputPorts];

		for (int i = 0; i < outputPorts; ++i) {
			outputSchemas[i] = getRuntimeOutputSchema(i);
		}
		
		try {
			scheduler = StdSchedulerFactory.getDefaultScheduler();
		} catch (SchedulerException e) {
			throw new StreamBaseException("Couldn't get an instance of the default Scheduler.", e);
		}
		try {
			scheduler.start();
		} catch (SchedulerException e) {
			throw new StreamBaseException("Couldn't start the default Scheduler.", e);
		}
	}

	/**
	 *  The shutdown method is called when the StreamBase server is in the process of shutting down.
	 */
	public void shutdown() {
		if(scheduler != null){
			try {
				scheduler.shutdown();
			} catch (SchedulerException e) {
				// nothing to do
			}
		}
	}

}
